import pathology_python_implementation as utility
import json
from datetime import datetime
import os
import xml.etree.ElementTree as et


class PathologyAnnotationXmlUtility:
    def __init__(self):
        self.root = None
        self.image_infos = []
        self.annotations = []
        self.categories = []
        self.parsed_xml_files = set()
        self.labels = {}
        self.license = {}
        self.dataset_info = {}
        self.image_counter = 0
        self.annotation_counter = 0

    def generate_coco_json(self, json_path):
        """
        Generate COCO JSON file.
        :param json_path: path to save the JSON file
        :return: None
        """
        self.dataset_info = {
            "description": "Copyright@DeepAgile Pathology XML in COCO format",
            "url": None,
            "year": datetime.today().strftime("%Y"),
            "version": "1.0",
            "contributor": "DeepAgile",
            "date_created": datetime.today().strftime("%Y/%m/%d"),
        }
        self.license = {
            "id": None,
            "name": None,
            "url": None
        }
        out_data = {
            "info": self.dataset_info,
            "license": self.license,
            "images": self.image_infos,
            "annotations": self.annotations,
            "categories": self.categories
        }
        json_str = json.dumps(out_data, indent=4)
        with open(json_path, "w") as f:
            f.write(json_str)

    def generate_yolo_txt(self, folder):
        """
        Generate YOLO TXT files.
        :param folder: the folder to save the TXT files
        :return: None
        """
        for idx in range(self.image_counter):
            file_name = self.image_infos[idx]["file_name"]
            image_width = float(self.image_infos[idx]["width"])
            image_height = float(self.image_infos[idx]["height"])
            image_id = self.image_infos[idx]["id"]
            annotations = [annotation for annotation in self.annotations if annotation["image_id"] == image_id]
            full_path = folder + "\\" + file_name + ".txt"
            with open(full_path, "w") as f:
                for annotation in annotations:
                    category_id = annotation["category_id"]
                    bbox = map(float, annotation["bbox"])
                    x_center, y_center, width, height = PathologyAnnotationUtility.transform_bbox_coco2yolo(image_width,
                                                                                                            image_height,
                                                                                                            bbox)
                    f.write(f"{category_id} {x_center:.2f} {y_center:.2f} {width:.2f} {height:.2f}\n")

    def generate_pascal_voc(self, folder):
        """
        Generate Pascal VOC XML files.
        :param folder: the folder to save the XML files
        :return: None
        """
        for idx in range(self.image_counter):
            file_name = self.image_infos[idx]["file_name"]
            image_width = self.image_infos[idx]["width"]
            image_height = self.image_infos[idx]["height"]
            image_id = self.image_infos[idx]["id"]
            annotations = [annotation for annotation in self.annotations if annotation["image_id"] == image_id]
            full_path = folder + "\\" + file_name + ".xml"
            annotation = et.Element("annotation")
            et.SubElement(annotation, "folder").text = ""
            et.SubElement(annotation, "filename").text = file_name
            size = et.SubElement(annotation, "size")
            et.SubElement(size, "width").text = str(image_width)
            et.SubElement(size, "height").text = str(image_height)
            et.SubElement(size, "depth").text = "3"

            for annotation_data in annotations:
                category_id = annotation_data["category_id"]
                bbox = map(float, annotation_data["bbox"])
                obj = et.SubElement(annotation, "object")
                et.SubElement(obj, "name").text = self.categories[category_id - 1]["name"]
                et.SubElement(obj, "pose").text = ""
                et.SubElement(obj, "truncated").text = "0"
                et.SubElement(obj, "difficult").text = "0"
                bndbox = et.SubElement(obj, "bndbox")
                x_min, y_min, x_max, y_max = list(
                    map(str, PathologyAnnotationUtility.transform_bbox_coco2pascal_voc(bbox)))
                et.SubElement(bndbox, "xmin").text = x_min
                et.SubElement(bndbox, "ymin").text = y_min
                et.SubElement(bndbox, "xmax").text = x_max
                et.SubElement(bndbox, "ymax").text = y_max

                # Add segmentation information (polygon)
                polygon = et.SubElement(obj, "polygon")
                for x, y in zip(annotation_data["segmentation"][0][0::2], annotation_data["segmentation"][0][1::2]):
                    pt = et.SubElement(polygon, "pt")
                    et.SubElement(pt, "x").text = str(x)
                    et.SubElement(pt, "y").text = str(y)

            tree = et.ElementTree(annotation)
            tree.write(full_path, encoding='utf-8', xml_declaration=True)

    def parse_xml_folder(self, folder):
        """
        Parse all XML files in a folder.
        :param folder: the folder containing XML files
        :return: None
        """
        for path in PathologyAnnotationUtility.get_xml_files(folder):
            self.parse_xml(path)

    def parse_xml(self, xml_path):
        """
        Parse an XML file.
        :param xml_path: the path to the XML file
        :return: None
        """
        image_info = utility.parse_image_information(xml_path)
        user_info = utility.parse_user_information(xml_path)
        self.add_image_info(image_info['file_name'], image_info['height'], image_info['width'], user_info['time'])
        self.get_rectangles(xml_path)
        self.get_polygons(xml_path)
        self.get_freehand_contours(xml_path)
        self.image_counter += 1

    def add_image_info(self, file_name, height, width, date_captured=None, image_license=None, flickr_url=None,
                       coco_url=None):
        """
        Add image information to the list.
        :param file_name: the name of the image
        :param height: the height of the image
        :param width: the width of the image
        :param date_captured: the date the image was captured
        :param image_license: the license of the image
        :param flickr_url: the Flickr URL of the image
        :param coco_url: the COCO URL of the image
        :return: None
        """
        self.image_infos.append({"id": self.image_counter, "height": height, "width": width, "file_name": file_name,
                                 "license": image_license, "flickr_url": flickr_url, "coco_url": coco_url,
                                 "date_captured": date_captured})

    def get_rectangles(self, xml_path):
        """
        Get all rectangles in an XML file.
        :param xml_path: the path to the XML file
        :return: None
        """
        rectangles = utility.parse_rectangle_annotations(xml_path)
        for rectangle in rectangles:
            points = list(map(float, rectangle['points'].split(",")))
            x_min, y_min = points[0], points[1]
            x_max, y_max = points[4], points[5]
            height = y_max - y_min
            width = x_max - x_min
            area = height * width
            bbox = [x_min, y_min, width, height]
            segmentation = [[x_min, y_min, x_max, y_min, x_max, y_max, x_min, y_max, x_min, y_min]]
            self.annotations.append({
                "id": self.annotation_counter,
                "category_id": self.get_label_id(rectangle['label']),
                "iscrowd": 0,
                "segmentation": segmentation,
                "image_id": self.image_counter,
                "area": area,
                "bbox": bbox,
            })

            self.annotation_counter += 1

    def get_polygons(self, xml_path):
        """
        Get all polygons in an XML file.
        :param xml_path: the path to the XML file
        :return: None
        """
        polygons = utility.parse_polygon_annotations(xml_path)
        for polygon in polygons:
            points = list(map(float, polygon['points'].split(",")))
            x_min, y_min, x_max, y_max = utility.calculate_bounding_box(points)
            height = y_max - y_min
            width = x_max - x_min
            area = height * width
            bbox = [x_min, y_min, width, height]
            segmentation = [points]
            segmentation[0].append(points[0])
            segmentation[0].append(points[1])
            self.annotations.append({
                "id": self.annotation_counter,
                "category_id": self.get_label_id(polygon['label']),
                "iscrowd": 0,
                "segmentation": segmentation,
                "image_id": self.image_counter,
                "area": area,
                "bbox": bbox,
            })

            self.annotation_counter += 1

    def get_freehand_contours(self, xml_path):
        """
        Get all freehand contours in an XML file.
        :param xml_path: the path to the XML file
        :return: None
        """
        freehand_contours = utility.parse_freehand_contour_annotations(xml_path)
        for freehand_contour in freehand_contours:
            points = list(map(float, freehand_contour['points'].split(",")))
            x_min, y_min, x_max, y_max = utility.calculate_bounding_box(points)
            height = y_max - y_min
            width = x_max - x_min
            area = height * width
            bbox = [x_min, y_min, width, height]
            segmentation = [points]
            segmentation[0].append(points[0])
            segmentation[0].append(points[1])
            self.annotations.append({
                "id": self.annotation_counter,
                "category_id": self.get_label_id(freehand_contour['label']),
                "iscrowd": 0,
                "segmentation": segmentation,
                "image_id": self.image_counter,
                "area": area,
                "bbox": bbox,
            })

            self.annotation_counter += 1

    def get_label_id(self, label):
        """
        Get the label ID, if the label does not exist, create a new one and return the ID.
        :param label: the label
        :return: the label ID
        """
        if label not in self.labels:
            self.labels[label] = len(self.labels) + 1
            self.categories.append({"id": self.labels[label], "name": label, "supercategory": None})
        return self.labels[label]

    def parse_label_list(self, xml_path):
        """
        Parse label list from an XML file.
        :param xml_path: the path to the XML file
        :return: label list
        """
        return utility.parse_label_list(xml_path)

    def parse_image_information(self, xml_path):
        """
        Parse image information from an XML file.
        :param xml_path: the path to the XML file
        :return: image information
        """
        return utility.parse_image_information(xml_path)

    def parse_user_information(self, xml_path):
        """
        Parse user information from an XML file.
        :param xml_path: the path to the XML file
        :return: user information
        """
        return utility.parse_user_information(xml_path)

    def parse_rectangle_annotations(self, xml_path):
        """
        Parse rectangles from an XML file.
        :param xml_path: the path to the XML file
        :return: rectangles
        """
        return utility.parse_rectangle_annotations(xml_path)

    def parse_line_annotations(self, xml_path):
        """
        Parse lines from an XML file.
        :param xml_path: the path to the XML file
        :return: lines
        """
        return utility.parse_line_annotations(xml_path)

    def parse_polygon_annotations(self, xml_path):
        """
        Parse polygons from an XML file.
        :param xml_path: the path to the XML file
        :return: polygons
        """
        return utility.parse_polygon_annotations(xml_path)

    def parse_circle_annotations(self, xml_path):
        """
        Parse circles from an XML file.
        :param xml_path: the path to the XML file
        :return: circles
        """
        return utility.parse_circle_annotations(xml_path)

    def parse_ruler_annotations(self, xml_path):
        """
        Parse rulers from an XML file.
        :param xml_path: the path to the XML file
        :return: rulers
        """
        return utility.parse_ruler_annotations(xml_path)

    def parse_ellipse_annotations(self, xml_path):
        """
        Parse ellipses from an XML file.
        :param xml_path: the path to the XML file
        :return: ellipses
        """
        return utility.parse_ellipse_annotations(xml_path)

    def parse_text_annotations(self, xml_path):
        """
        Parse texts from an XML file.
        :param xml_path: the path to the XML file
        :return: texts
        """
        return utility.parse_text_annotations(xml_path)

    def parse_freehand_contour_annotations(self, xml_path):
        """
        Parse freehand contours from an XML file.
        :param xml_path: the path to the XML file
        :return: freehand contours
        """
        return utility.parse_freehand_contour_annotations(xml_path)

    def parse_label_list(self, xml_path):
        """
        Parse label list from an XML file.
        :param xml_path: the path to the XML file
        :return: label list
        """
        return utility.parse_label_list(xml_path)

    def get_all_annotations(self, xml_path):
        """
        Get all annotations from an XML file.
        :param xml_path: the path to the XML file
        :return: all annotations
        """
        return utility.get_all_annotations(xml_path)

    @staticmethod
    def transform_bbox_coco2yolo(image_width, image_height, bounding_box):
        """
        Transform bounding box from COCO format to YOLO format.
        :param image_width: width of the image
        :param image_height: height of the image
        :param bounding_box: bounding box in COCO format
        :return: bounding box in YOLO format
        """
        x, y, width, height = bounding_box
        x_center = x + width / 2
        y_center = y + height / 2
        x_center /= image_width
        y_center /= image_height
        width /= image_width
        height /= image_height
        return x_center, y_center, width, height

    @staticmethod
    def transform_bbox_yolo2coco(image_width, image_height, bounding_box):
        """
        Transform bounding box from YOLO format to COCO format.
        :param image_width: width of the image
        :param image_height: height of the image
        :param bounding_box: bounding box in YOLO format
        :return: bounding box in COCO format
        """
        x_center, y_center, width, height = bounding_box
        x = x_center * image_width - width / 2
        y = y_center * image_height - height / 2
        width *= image_width
        height *= image_height
        return x, y, width, height

    @staticmethod
    def transform_bbox_coco2pascal_voc(bounding_box):
        """
        Transform bounding box from COCO format to Pascal VOC format.
        :param bounding_box: bounding box in COCO format
        :return: bounding box in Pascal VOC format
        """
        x, y, width, height = bounding_box
        x_min = x
        y_min = y
        x_max = x + width
        y_max = y + height
        return x_min, y_min, x_max, y_max

    @staticmethod
    def transform_bbox_pascal_voc2coco(bounding_box):
        """
        Transform bounding box from Pascal VOC format to COCO format.
        :param bounding_box: bounding box in Pascal VOC format
        :return: bounding box in COCO format
        """
        x_min, y_min, x_max, y_max = bounding_box
        x = x_min
        y = y_min
        width = x_max - x_min
        height = y_max - y_min
        return x, y, width, height

    @staticmethod
    def get_xml_files(folder):
        return [os.path.join(folder, f) for f in os.listdir(folder) if f.endswith(".xml")]
